-- FILE:     DFS.E
-- PROJECT:  DOS-File-System library routines with full Long File Name support.
-- PURPOSE:  The elementary drive/directory/file/path/critical-error routines.
-- AUTHOR:   Shian Lee
-- VERSION:  1.00  December/18/2012
-- LANGUAGE: Euphoria version 3.1.1 (http://www.RapidEuphoria.com)
-- PLATFORM: DOS32
-- LICENCE:  Free. Use at your own risk.
-- NOTE:     * This library comes with "DFS.DOC" standard user manual.
--           * basic programming info is commented above each routine.
--           * this file assumes a constant DTA - see DFS_DTA_ADDRESS.
--           * local naming and assumptions for all routines:
--             - pn = absolute or relative, short or long, pathname string.
--             - drive-letters (A..Z) always returned upper-cased, e.g. 'C'.
-- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
-- HISTORY:  DFS 1.00, December/18/2012: Tested on FreeDOS 1.1. First version.
-----------------------------------------------------------------------------



include MACHINE.E       -- include Euphoria 3.1 Machine library routines



--bookmark: LOCAL CONSTANTS & TYPES; ========================================

-- PURPOSE: make code more clear by using constants.
constant
    -- boolean (0/1) constants:
    TRUE = (1 = 1),         -- 1 (true)
    FALSE = not TRUE,       -- 0 (false)
    -- other common names for booleans:
    SET = TRUE,             -- 1 (set)
    GET = FALSE,            -- 0 (get)
    CLEAR = FALSE,          -- 0 (clear)

    -- commonly used values:
    NULL = 0,               -- 0  (NULL is the name of ASCII-0 character).
    FAIL = -1,              -- -1 (some routines return -1 on error/failure).
    EMPTY = {},             -- {} (empty string or sequence).

    -- mask to upper('a'..'z'/'A'..'Z') drive-letter, by clearing bit 5.
    -- e.g.: i = and_bits('c', UPPER)  --> i is 'C'
    UPPER = #DF             -- #DF = 223 = 0b_1101_1111


-- PURPOSE: test if b is boolean (0 or 1).
type boolean(integer b)
    return b = FALSE or b = TRUE
end type

-- PURPOSE: test if l is a drive letter (A..Z/a..z).
type drive_letter(integer l)
    return (l >= 'A' and l <= 'Z') or (l >= 'a' and l <= 'z')
end type

-- PURPOSE: test if l is a drive letter (A..Z/a..z) or @ (@=default drive).
type xDrive_letter(integer l)
    return (l >= '@' and l <= 'Z') or (l >= 'a' and l <= 'z')
end type

-- PURPOSE: test if l is 'n'/'f'/'d'/'v' (normal/file/dir/volume wdir() mode).
type nfdv_char(integer l)
    return l = 'n' or l = 'f' or l = 'd' or l = 'v'
end type



--bookmark: FIX PATH ROUTINES; ==============================================

-- PURPOSE: trim invalid characters in absolute/relative DOS pathnames:
--           1. leading/trailing spaces/control-characters/quotation-marks.
--           2. backslash in the end of pathname (unless it's a root-dir).
--           3. dot(s) in the end of filename without extension.
-- NOTE:    all global routines in DFS.E doing trim_path() automatically.
-- RETURN:  string; (the invalid characters are removed).
-- EXAMPLE: s = trim_path("  C:\\PATHNAME...\\  ")   --> s is "C:\PATHNAME"
--          s = trim_path("\\" & 13)                 --> s is "\"
--          s = trim_path("..\\ ")                   --> s is ".."
--          s = trim_path(" \" C:\\Long Name\\ \" ") --> s is "C:\Long Name"
-- SEEALSO: slash_path
global function trim_path(sequence pn)
    integer l, t                            -- length, temp-variable

    l = length(pn)                          -- initialize l (length)

    -- 1. trim spaces, control-characters and quotation marks:
    while l and (pn[l] <= 32 or pn[l] = '"') do
	l -= 1                              -- trim right
	pn = pn[1..l]                       --> faster inside loop since path
    end while                               --   is usually trimmed already.
    while l and (pn[1] <= 32 or pn[1] = '"') do
	pn = pn[2..l]                       -- trim left
	l -= 1
    end while

    -- 2. trim invalid '\' from end of pn if is not "\" or "x:\" (root):
    if l > 1 and pn[l] = '\\' and pn[l - 1] != ':' then
	l -= 1
	pn = pn[1..l]                       -- trim invalid backslash
    end if

    -- 3. trim invalid dot(s) from end of filename without extension:
    t = l                                   -- remember current length
    while l > 1 and pn[l] = '.' do
	l -= 1                              -- count down dots
    end while
    if l < t and pn[l] != '\\' and pn[l] != ':' and pn[l] != '.' then
	pn = pn[1..l]                       -- trim invalid dot(s)
 -- else
 --     l = t                               -- dot(s) is a relative-dir;
    end if                                  --  so restore current length.

    return pn
end function



-- PURPOSE: end pn with backslash so it's possible to append a filename to it.
-- NOTE:    use trim_path() to remove invalid backslash from pn.
-- RETURN:  string; pn ends with '\' or ':'. if pn was "" then return ".\"
-- EXAMPLE: s = slash_path(" C:\\DIRNAME. ")   --> s is "C:\DIRNAME\"
--          s = slash_path(" C:\\ ")           --> s is "C:\"
-- SEEALSO: trim_path
global function slash_path(sequence pn)
    integer l                               -- pn length

    pn = trim_path(pn)                      -- trim invalid characters.
    l = length(pn)                          -- get length of pn

    if l then                               -- if pn is not empty:
	if pn[l] = '\\' or pn[l] = ':' then -- cannot add \ after '\' or ':'
	    return pn                       -- do not add \ to pn.
	else
	    return pn & '\\'                -- add \ to pn.
	end if
    else                                    -- if pn is empty (""):
	return ".\\"                        -- assumes current-dir (".")
    end if
end function



-- PURPOSE: trim a null-terminated-string (ASCIIZ).
-- DETAILS: asciiz is a string with or without a NULL character.
-- RETURN:  string;
-- EXAMPLE: s = trim_asciiz("C:\\TEMP" & NULL & "XYZ") --> s is "C:\TEMP"
--          s = trim_asciiz("C:\\TEMP")             --> s is "C:\TEMP"
function trim_asciiz(sequence asciiz)
    integer z

    z = find(NULL, asciiz)                  -- find NULL (ASCII-0) in string
    if z then
	return asciiz[1..z - 1]             -- return the actual ASCII string
    else
	return asciiz                       -- return unchanged ASCII string
    end if
end function



--bookmark: PARSE PATH ROUTINES; ============================================

-- PURPOSE: return the filename portion of pn (pn may exist or not).
-- RETURN:  string; EMPTY/"filename" (no-filename/filename-portion)
-- EXAMPLE: s = file_name(" C:\\DIR1\\FILE.DOC ")   --> s is "FILE.DOC"
--          s = file_name(" C:\\DIR1\\ ")           --> s is "DIR1"
--          s = file_name(" C:\\DIR1\\.. ")         --> s is ""
--          s = file_name(" C:\\ ")                 --> s is ""
-- SEEALSO: filebase_name, fileext_name, real_path
global function file_name(sequence pn)
    integer l                               -- pn length

    pn = trim_path(pn)                      -- trim invalid characters.
    l = length(pn)                          -- get length of pn

    -- relative dir "", ".", "..", "...", "\" or ":" is not a filename:
    if l = 0 or pn[l] = '.' or pn[l] = '\\' or pn[l] = ':' then
	return EMPTY                        -- filename not exist (only dir)
    else
	for i = l to 1 by -1 do             -- search for filename
	    if pn[i] = '\\' or pn[i] = ':' then
		return pn[i + 1..l]         -- filename found
	    end if
	end for
	return pn                           -- filename = pn
    end if
end function



-- PURPOSE: return the base-filename portion of pn (pn may exist or not).
-- RETURN:  string; EMPTY/"base-filename" (no-filename/base-filename-portion)
-- EXAMPLE: s = filebase_name("C:\\TMP\\LONG.NAME.DOC") --> s is "LONG.NAME"
-- SEEALSO: file_name, fileext_name
global function filebase_name(sequence pn)
    sequence file

    file = file_name(pn)                    -- file_name() will trim_path(),

    -- LFN can have many dots, so find the right-most dot:
    for i = length(file) to 1 by -1 do
	if file[i] = '.' then
	    return file[1..i - 1]           -- return base-filename
	end if
    end for

    return file                             -- base-filename is filename
end function



-- PURPOSE: return the filename-extension portion of pn (pn may exist or not).
-- RETURN:  string; EMPTY/"filename-extension" (no-filename/extension-portion).
-- EXAMPLE: s = fileext_name("C:\\TMP\\LONG.NAME.DOC") --> s is "DOC"
-- SEEALSO: file_name, filebase_name
global function fileext_name(sequence pn)
    sequence file

    file = file_name(pn)                    -- file_name() will trim_path(),

    -- LFN can have many dots, so find the right-most dot:
    for i = length(file) to 1 by -1 do
	if file[i] = '.' then
	    return file[i + 1..$]           -- return filename-extension
	end if
    end for

    return EMPTY                            -- extension not exists.
end function



-- PURPOSE: return the dir portion of pn (pn may exist or not).
-- RETURN:  string; EMPTY/"...." (no-dir/dir-portion)
-- EXAMPLE: s = dir_name(" C:\\DIR1\\FILE.DOC ")   --> s is "C:\DIR1"
--          s = dir_name(" FILE.DOC ")             --> s is ""
-- SEEALSO: drive_name, file_name, real_dir
global function dir_name(sequence pn)
    pn = trim_path(pn)                      -- trim invalid characters.

    return trim_path(                       -- trim trailing '\' from dir
	pn[1..$ - length(file_name(pn))]
    )
end function



-- PURPOSE: return the drive portion of pn (pn may exist or not).
-- RETURN:  string; EMPTY/"x" (no-drive/drive-portion)
-- EXAMPLE: s = drive_name(" C:\\DIR1\\FILE.DOC ") --> s is "C"
--          s = drive_name(" \\DIR1\\FILE.DOC ")   --> s is ""
-- SEEALSO: dir_name, file_name, real_drive
global function drive_name(sequence pn)
    pn = trim_path(pn)                      -- trim invalid characters.

    if length(pn) >= 2 and pn[2] = ':' then
	return {pn[1]}                      -- return "x"
    else
	return EMPTY                        -- drive not exist or invalid
    end if
end function



--bookmark: LFN SPECIFIC ROUTINES; ==========================================

-- PURPOSE: return existing fullname in long or short [8.3] format (LFN only).
--          or check if: LFN-supported = sequence(lfn_real_path(".", TRUE))
-- DETAILS: mode is: FALSE = get short [8.3] path. TRUE = get long path.
-- RETURN:  object; integer FAIL/string fullname (error/successful).
-- EXAMPLE: x = lfn_real_path(".", TRUE)  --> x is FAIL if LFN *not* supported.
--          x = lfn_real_path(".", TRUE)  --> x may be "C:\Temp Directory", etc.
--          x = lfn_real_path(".", FALSE) --> x may be "C:\TEMPDI~1", etc.
-- SEEALSO: real_short_path, real_long_path, allow_lfn, real_path
function lfn_real_path(sequence pn, boolean mode)
    sequence r                              -- list of register values
    integer ad_pn, ad_pn2                   -- low memory addresses
    sequence pn2                            -- full pathname (output string)
    integer len_pn2                         -- full pathname maximum-length

    -- get value for ASCIIZ full pathname maximum-length:
    if mode then
	len_pn2 = 260 + 1                   -- long=260-bytes (usually...)
    else
	len_pn2 = 128 + 1                   -- short=128-bytes (theoretically)
    end if

    pn = trim_path(pn)                      -- trim invalid characters.

    ad_pn = allocate_low(length(pn) + 1)    -- low memory to hold ASCIIZ pn
    if not ad_pn then                       -- memory can't be allocated?
	return FAIL
    else
	ad_pn2 = allocate_low(len_pn2)      -- low memory to hold ASCIIZ pn2
	if not ad_pn2 then                  -- memory can't be allocated?
	    free_low(ad_pn)                 -- free the pn buffer
	    return FAIL
	end if
    end if

    poke(ad_pn, pn & NULL)                  -- poke ASCIIZ pn to buffer
    poke(ad_pn2, NULL)                      -- poke NULL to pn2 for compare
    r = repeat(CLEAR, REG_LIST_SIZE)        -- initialize registers list
    r[REG_AX] = #7160                       -- AX=#7160 (LFN) - TRUENAME function
    r[REG_CX] = #8000 + (mode and TRUE) + 1 -- CL=#01/#02 - Get Short/Long Pathname sub-function
					    -- CH=#80 - return a path containing the SUBSTed drive letter
					    -- (if CH=#00 - return a path containing true path for a SUBSTed drive letter)
    r[REG_DS] = floor(ad_pn / 16)           -- DS:SI=segment:offset for
    r[REG_SI] = remainder(ad_pn, 16)        --  ASCIIZ pn (input string)
    r[REG_ES] = floor(ad_pn2 / 16)          -- ES:DI=segment:offset for
    r[REG_DI] = remainder(ad_pn2, 16)       --  ASCIIZ pn2 (output string)
    r = dos_interrupt(#21, r)               -- call DOS interrupt #21
    pn2 = peek({ad_pn2, len_pn2})           -- get pn2 from output buffer
    free_low(ad_pn)                         -- free pn buffer
    free_low(ad_pn2)                        -- free pn2 buffer

    -- Carry Flag set on error only by DOS7+ for LFN functions, so compare
    -- buffer to make sure that LFN functions are supported by the OS
    -- (note that on any error ES:DI buffer remains unchanged):
    if pn2[1] = NULL then                   -- on error first byte is NULL
	return FAIL
    else                                    -- successful
	return trim_asciiz(pn2)
    end if
end function



-- PURPOSE: return fullname for existing pn in short format (8.3) (LFN only).
-- RETURN:  object; integer FAIL/string fullname (error/successful).
-- EXAMPLE: x = real_short_path(".")     --> get current-dir in short format
-- SEEALSO: real_long_path, allow_lfn
global function real_short_path(sequence pn)
    return lfn_real_path(pn, FALSE)         -- FALSE = get short pathname
end function



-- PURPOSE: return fullname for existing pn in long format (LFN only).
-- RETURN:  object; integer FAIL/string fullname (error/successful).
-- EXAMPLE: x = real_long_path(".")     --> get current-dir in long format
-- SEEALSO: real_short_path, allow_lfn
global function real_long_path(sequence pn)
    return lfn_real_path(pn, TRUE)          -- TRUE = get long pathname
end function



-- PURPOSE: keeps the status of LFN support in DFS.E (for *all* routines).
-- NOTE:    initialized on start using allow_lfn(TRUE).
-- RETURN:  boolean; FALSE/TRUE (LFN disabled/LFN enabled).
--           FALSE = use regular interrupts compatible with MS-DOS 3.2+.
--           TRUE  = use new interrupts compatible with MS-DOS 7.0+ (Win95+).
-- SEEALSO: allow_lfn
boolean DFS_lfn



-- PURPOSE: enable or disable long-file-names (LFN) support in DFS.E.
-- DETAILS: allow is FALSE/TRUE (disable/enable LFN support in DFS.E).
-- OUTPUT:  DFS_lfn
-- EXAMPLE: allow_lfn(TRUE)    --> enable LFN support (if supported by OS).
-- SEEALSO: check_lfn, DFS_lfn
global procedure allow_lfn(boolean allow)
    if allow then
	-- test if LFN is *currently* supported by the OS driver:
	DFS_lfn = sequence(lfn_real_path(".", TRUE)) -- TRUE = get long name
    else
	DFS_lfn = FALSE
    end if
end procedure



-- PURPOSE: enable LFN support on start by default (if supported by the OS).
allow_lfn(TRUE)



-- PURPOSE: return current status of long-file-names (LFN) support in DFS.E.
-- INPUT:   DFS_lfn
-- RETURN:  boolean; FALSE/TRUE (LFN support disabled/enabled).
-- EXAMPLE: i = check_lfn()
-- SEEALSO: allow_lfn, DFS_lfn
global function check_lfn()
    return DFS_lfn
end function



--bookmark: wdir() FUNCTION ROUTINES; =======================================

-- PURPOSE: get the current disk transfer area (DTA) address.
-- NOTE:    DFS.E is using the DTA for wdir() function, Int #21 AH=#4E/#4F.
-- RETURN:  integer; current DTA low memory address.
-- EXAMPLE: i = get_dta_address()
-- SEEALSO: DFS_DTA_ADDRESS, wdir_entry, wdir
function get_dta_address()
    sequence r                              -- list of register values

    r = repeat(CLEAR, REG_LIST_SIZE)        -- initialize registers list
    r[REG_AX] = #2F00                       -- AH=#2F - Get DTA address function
    r = dos_interrupt(#21, r)               -- call DOS interrupt #21
    return (r[REG_ES] * 16) + r[REG_BX]     -- ES:BX=segment:offset of DTA
end function



-- PURPOSE: assuming that Euphoria/C is setting a safe DTA, i.e. that will not
--          cause a segment wrap, and using a *constant* DTA address for dir()
--          function - DFS.E will also use this address for wdir() function.
--          note that wdir() needs 43-bytes buffer in the DTA.
-- NOTICE:  * if you set the DTA, you MUST restore DFS_DTA_ADDRESS of probably
--            128-bytes buffer, or even 256-bytes (as initially was set by C).
--          * DO NOT set the DTA before DFS.E initializes DFS_DTA_ADDRESS!
--            since DFS.E assumes that the current DTA is set by Euphoria/C.
-- RETURN:  integer; current DTA low memory address (previously set by C).
-- EXAMPLE: see wdir_entry() function.
-- SEEALSO: get_dta_address, wdir_entry, wdir
global constant
    DFS_DTA_ADDRESS = get_dta_address()



-- PURPOSE: converts attributes byte to attributes string in dir()'s format.
-- DETAILS: attr is dir-entry-attributes byte (6-bits in DOS format).
-- RETURN:  string; entry attributes in dir()'s format (D_ATTRIBUTES).
-- EXAMPLE: s = entry_attrib_to_string(#21)  --> s is "ra"
-- SEEALSO: wdir_entry, entry_date_time_to_sequence, attrib
constant
    -- (this constant is used also by attrib() function):
    ATTR_MASK = {{#01, #02, #04, #08, #10, #20} , "rhsvda"}
function entry_attrib_to_string(integer attr)
    sequence s
						    --| Attributes-Byte
    s = EMPTY   -- initialize to no-attributes.     --| bits (DOS format):
						    --| ------------------
    for i = 1 to 6 do                               --| 0 = 'r' read-only
	if and_bits(attr, ATTR_MASK[1][i]) then     --| 1 = 'h' hidden
	    s &= ATTR_MASK[2][i]                    --| 2 = 's' system
	end if                                      --| 3 = 'v' volume label
    end for                                         --| 4 = 'd' directory
    return s                                        --| 5 = 'a' archive
end function



-- PURPOSE: converts date-time words to date-time sequence in dir()'s format.
-- NOTE:    this is the opposite function of sequence_to_entry_date_time().
-- DETAILS: d is dir-entry-date (16-bits in DOS format).
--          t is dir-entry-time (16-bits in DOS format).
--          * note that d and t must be 32-bits atoms.
-- RETURN:  sequence; entry date-time in dir()'s format {D_YEAR..D_SECOND}.
-- EXAMPLE: s = entry_date_time_to_sequence(#4155, #98137A90)
--           --> s is {2012,10,21,15,20,32}.
-- SEEALSO: wdir_entry, entry_attrib_to_string
function entry_date_time_to_sequence(atom d, atom t)

    -- manipulating only the low 16-bits of date and time:
    d = and_bits(d, #FFFF)
    t = and_bits(t, #FFFF)

    return {
	-- date word:           -- index:     bits (DOS format):
	floor(d / #200) + 1980, -- 4.D_YEAR   9-15 =year (relative to 1980)
	and_bits(d / #20, #F),  -- 5.D_MONTH  5-8  =month (1..12)
	and_bits(d, #1F),       -- 6.D_DAY    0-4  =day (1..31)
	-- time word:
	floor(t / #800),        -- 7.D_HOUR   11-15=hours (0..23)
	and_bits(t / #20, #3F), -- 8.D_MINUTE 5-10 =minutes (0..59)
	and_bits(t, #1F) * 2    -- 9.D_SECOND 0-4  =2-second increments (0..29)
    }
end function



-- PURPOSE: converts date-time sequences in dir()'s format to date-time words.
-- NOTE:    this is the opposite function of entry_date_time_to_sequence().
-- DETAILS: d is dir()'s {D_YEAR, D_MONTH, D_DAY}.  --> see Euphoria's dir().
--          t is dir()'s {D_HOUR, D_MINUTE, D_SECOND}.
-- RETURN:  sequence; {integer date, integer time}
--           * date & time are 16-bits words in DOS file-entry format.
-- EXAMPLE: s = sequence_to_entry_date_time({2012,10,21}, {15,20,32})
--           --> s is {#4155, #7A90}.
-- SEEALSO: entry_date_time_to_sequence, used-by: date_file
function sequence_to_entry_date_time(sequence d, sequence t)
    return {
	-- date word:                       -- index:     bits (DOS format):
	and_bits(d[1] - 1980, #7F) * #200 + -- 4.D_YEAR   9-15 =year (relative to 1980)
	and_bits(d[2], #F) * #20 +          -- 5.D_MONTH  5-8  =month (1..12)
	and_bits(d[3], #1F),                -- 6.D_DAY    0-4  =day (1..31)
	-- time word:
	and_bits(t[1], #1F) * #800 +        -- 7.D_HOUR   11-15=hours (0..23)
	and_bits(t[2], #3F) * #20 +         -- 8.D_MINUTE 5-10 =minutes (0..59)
	and_bits(t[3] / 2, #1F)             -- 9.D_SECOND 0-4  =2-second increments (0..29)
    }
end function



-- PURPOSE: return a single dir-entry-info in dir()'s format.
-- DETAILS: ad_fdr is the low memory address of a 317-byte FindData record (LFN).
-- INPUT:   DFS_lfn
-- RETURN:  sequence; a single dir()'s entry.
-- EXAMPLE: s = wdir_entry(ad_fdr)
--           --> s may be {"fred", "ra", 2350, 1994, 1, 22, 17, 22, 40}
-- SEEALSO: DFS_DTA_ADDRESS, used-by: wdir
function wdir_entry(integer ad_fdr)

    -- LFN - the search results returned in the 317-byte FindData record:
    if DFS_lfn then
	return {
	    trim_asciiz(                    -- 1.D_NAME (LFN),
		peek({ad_fdr + #2C, 260})
	    ),
	    entry_attrib_to_string(         -- 2.D_ATTRIBUTES,
		peek(ad_fdr)
	    ),
	    peek4u(ad_fdr + #1C) * #100000000 + -- 3.D_SIZE, (high 32 bits)
	    peek4u(ad_fdr + #20)                --           (low 32 bits)
	} &
	entry_date_time_to_sequence(
	    peek4u(ad_fdr + #16),           -- 4.D_YEAR,5.D_MONTH,6.D_DAY,
	    peek4u(ad_fdr + #14)            -- 7.D_HOUR,8.D_MINUTE,9.D_SECOND
	)

    -- the search results returned in current disk transfer area (DTA):
    else
	return {
	    trim_asciiz(                    -- 1.D_NAME,
		peek({DFS_DTA_ADDRESS + #1E, 13})
	    ),
	    entry_attrib_to_string(         -- 2.D_ATTRIBUTES,
		peek(DFS_DTA_ADDRESS + #15)
	    ),
	    peek4u(DFS_DTA_ADDRESS + #1A)   -- 3.D_SIZE,
	} &
	entry_date_time_to_sequence(
	    peek4u(DFS_DTA_ADDRESS + #18),  -- 4.D_YEAR,5.D_MONTH,6.D_DAY,
	    peek4u(DFS_DTA_ADDRESS + #16)   -- 7.D_HOUR,8.D_MINUTE,9.D_SECOND
	)
    end if
end function



-- PURPOSE: free the Find-First-File handle for reuse (LFN only).
-- DETAILS: handle is returned by AX=#714E (LFN - Find First File function).
-- EXAMPLE: terminate_dir_search(1)
-- SEEALSO: wdir
procedure terminate_dir_search(integer handle)
    sequence r                              -- list of register values

    r = repeat(CLEAR, REG_LIST_SIZE)        -- initialize registers list
    r[REG_AX] = #71A1                       -- AX=#71A1 (LFN) - FindClose function
    r[REG_BX] = handle                      -- BX=filefind handle (from AX=#714E)
    r = dos_interrupt(#21, r)               -- call DOS interrupt #21
					    -- ( Carry Flag set on error )
    -- < the following code is for debugging only >:
    -- printf(1, "\nDEBUG: filefind handle %d closed? %d\n", {handle, not and_bits(r[REG_FLAGS], 1)})
end procedure



-- PURPOSE: return dir info in Euphoria's dir() format (wdir = wildcards-dir).
-- NOTICE:  this function assumes that the DTA is previously set for 43-bytes+.
-- DETAILS: pn may have two formats: "pn" or {"pn", mode}.
--          pn is pathname of file, dir or volume-label. you must specify
--           wildcards, * ?, to return the *content* of directory.
--          mode is: 'n' = find files and directories (normal search).
--                   'f' = find only files.
--                   'd' = find only directories.
--                   'v' = find volume label (only in the root-dir).
--           * the default mode is 'n' (normal) if mode is not specified.
-- INPUT:   DFS_lfn
-- RETURN:  object; integer FAIL/sequence dir()'s-info (error/successful).
--           * see Euphoria's dir() for complete details.
-- EXAMPLE: x = wdir("C:\\TEMP_DIR\\*.*")
--           --> get the *content* of TEMP_DIR using wildcards (* and ?).
--          x = wdir({"C:\\Temp Directory\\fred", 'f'})
--           --> x may be {{"fred", "ra", 2350, 1994, 1, 22, 17, 22, 40}}
-- SEEALSO: volume_label, attrib, file_exists
global function wdir(sequence pn)
    sequence r                              -- list of register values
    integer ad_pn, ad_fdr                   -- low memory addresses
    integer segment_fdr, offset_fdr         -- FindData record seg:offs (LFN)
    integer allow_attr, req_attr            -- search attribute bytes
    integer handle                          -- filefind handle (from AX=#714E)
    sequence entry                          -- a single dir() entry
    sequence info                           -- one or more dir() entries
    integer findOnlyDirs                    -- find only directories flag
    nfdv_char mode                          -- search mode ('n'/'f'/'d'/'v')

    -- set specified or default search mode:
    if length(pn) and sequence(pn[1]) then
	mode = pn[2]                        -- mode specified ({"pn", mode}).
	pn = pn[1]
    else
	mode = 'n'                          -- use default mode 'n' ("pn").
    end if

    pn = trim_path(pn)                      -- trim invalid characters.

    if not length(pn) then                  -- pn not specified?
	pn = "."                            --  then pn is current-dir (".").
    end if

    findOnlyDirs = (mode = 'd')             -- clear/set flag (speedup loop)

    -- set search attribute byte:
    if mode = 'n' then                      -- files & directories (normal):
	allow_attr = #37                    --  0b_0011_0111 = #37 = "rhsda"
	  req_attr = #00                    --  clear required-attribute (LFN)
    elsif mode = 'f' then                   -- only files:
	allow_attr = #27                    --  0b_0010_0111 = #27 = "rhsa"
	  req_attr = #00                    --  clear required-attribute (LFN)
    elsif mode = 'd' then                   -- only directories:
	allow_attr = #37                    --  0b_0011_0111 = #37 = "rhsda"
	  req_attr = #10                    --  0b_0001_0000 = #10 = "d" (LFN)
    elsif mode = 'v' then                   -- only a single volume label:
	allow_attr = #08                    --  0b_0000_1000 = #08 = "v"
	  req_attr = #08                    --  0b_0000_1000 = #08 = "v" (LFN)
    end if


    -- * start of Find First File function * --
    pn &= NULL                              -- ASCII --> ASCIIZ
    ad_pn = allocate_low(length(pn))        -- low memory buffer to hold pn
    if not ad_pn then                       -- memory can't be allocated?
	return FAIL
    elsif DFS_lfn then
	ad_fdr = allocate_low(317)          -- low memory to hold 317-byte FindData record
	if not ad_fdr then                  -- memory can't be allocated?
	    free_low(ad_pn)                 -- free the pn buffer
	    return FAIL
	else
	    segment_fdr = floor(ad_fdr / 16)    -- FindData record segment:offset
	    offset_fdr = remainder(ad_fdr, 16)  --  (to speedup loop).
	end if
    else
	ad_fdr = 0                          -- just init if DFS_lfn = FALSE
    end if

    poke(ad_pn, pn)                         -- poke ASCIIZ pn to buffer
    r = repeat(CLEAR, REG_LIST_SIZE)        -- initialize registers list
    if DFS_lfn then                         -- Find First File function:
	r[REG_AX] = #714E                   --  AX=#714E (LFN)
	r[REG_CX] =                         --  CL=allowable-attributes mask
	    (req_attr * #100) + allow_attr  --  CH=required-attributes mask
	r[REG_SI] = #0001                   --  SI=#0001 - DOS date/time format
	r[REG_ES] = segment_fdr             --  ES:DI=segment:offset of
	r[REG_DI] = offset_fdr              --   317-byte FindData record
    else
	r[REG_AX] = #4E00                   --  AH=#4E   (regular)
	r[REG_CX] = allow_attr              --  CX=search attribute
    end if
    r[REG_DS] = floor(ad_pn / 16)           -- DS:DX=segment:offset of
    r[REG_DX] = remainder(ad_pn, 16)        --  ASCIIZ pn
    r = dos_interrupt(#21, r)               -- call DOS interrupt #21

    -- Carry Flag set on error:
    if and_bits(r[REG_FLAGS], 1) then
	free_low(ad_pn)                     -- free pn buffer
	if DFS_lfn then
	    free_low(ad_fdr)                -- free FindData record buffer
	end if
	return FAIL                         -- matching file not found

    -- return a single entry or volume-label (not the directory content):
    elsif not (find('*', pn) or find('?', pn)) or mode = 'v' then
	entry = wdir_entry(ad_fdr)          -- get a single dir() entry
	free_low(ad_pn)                     -- free pn buffer
	if DFS_lfn then
	    free_low(ad_fdr)                -- free FindData record buffer
	    terminate_dir_search(r[REG_AX]) -- AX=filefind handle
	end if
	if findOnlyDirs and not find('d', entry[2]) then
	    return FAIL                     -- file is found - not dir
	else
	    return {entry}                  -- matching entry found
	end if

    -- AX=filefind handle (needed to continue LFN search):
    elsif DFS_lfn then
	handle = r[REG_AX]                  -- filefind handle (from AX=#714E)
    end if


    -- * start of Find Next File function * --
    info = EMPTY                            -- initialize directory content
    while not and_bits(r[REG_FLAGS], 1) do  -- Carry Flag set on error

	entry = wdir_entry(ad_fdr)          -- get a single dir() entry
	if findOnlyDirs and not find('d', entry[2]) then
	    -- don't add this entry.        -- file is found - not dir
	else
	    info = append(info, entry)      -- add entry to directory content
	end if

	r = r and CLEAR                     -- clear registers list
	if DFS_lfn then                     -- Find Next File function:
	    r[REG_AX] = #714F               --  AX=#714F (LFN)
	    r[REG_BX] = handle              --  BX=filefind handle (from AX=#714E)
	    r[REG_SI] = #0001               --  SI=#0001 - DOS date/time format
	    r[REG_ES] = segment_fdr         --  ES:DI=segment:offset of
	    r[REG_DI] = offset_fdr          --   317-byte FindData record
	else
	    r[REG_AX] = #4F00               --  AH=#4F   (regular)
	end if
	r = dos_interrupt(#21, r)           -- call DOS interrupt #21
    end while

    free_low(ad_pn)                         -- free pn buffer
    if DFS_lfn then
	free_low(ad_fdr)                    -- free FindData record buffer
	terminate_dir_search(handle)
    end if

    if equal(info, EMPTY) then              -- if mode = 'd' and only files
	return FAIL                         --  found then info is EMPTY.
    else
	return info                         -- matching entry(s) found
    end if
end function



--bookmark: CRITICAL ERROR HANDLER ROUTINES; ================================

-- PURPOSE: check if a critical error occurred since the last call.
-- NOTE:    calling allow_error() always reset to FALSE (no-error).
-- INPUT:   INT24_FLAG_ADDRESS
-- RETURN:  boolean; FALSE/TRUE
--                   -> FALSE = no critical error occurred since last call.
--                   -> TRUE  = a critical error occurred since last call.
-- EXAMPLE: i = check_error()
-- SEEALSO: allow_error

-- setup the 4-bytes critical error flag:
constant INT24_FLAG_ADDRESS = allocate(4)   -- (unlikely to return 0...)
lock_memory(INT24_FLAG_ADDRESS, 4)          -- lock buffer forever
poke4(INT24_FLAG_ADDRESS, CLEAR)            -- clear flag (no error)

global function check_error()
    -- a critical error occurred since last check (flag is set)?
    if peek4u(INT24_FLAG_ADDRESS) then
	poke4(INT24_FLAG_ADDRESS, CLEAR)    -- clear flag (no error)
	return TRUE
    -- no critical error occurred since last check:
    else
	return FALSE
    end if
end function



-- PURPOSE: load DFS-critical-error-handler to memory and lock it forever.
-- NOTE:    this function should be called by allow_error() only once!
-- CREDIT:  * Euphoria 3.1 - demo\dos32\hardint.ex
--            (handling interrupt in either real mode or protected mode).
-- INPUT:   INT24_FLAG_ADDRESS
-- OUTPUT   Code_Segment_Address, INT24_dfs_address
-- EXAMPLE: setup_DFS_critical_error_handler()  --> (that's all).
-- SEEALSO: allow_error

atom INT24_dfs_address                      -- Int #24 DFS-handler address.
    INT24_dfs_address = FALSE               -- FALSE=DFS-handler not loaded yet.
atom Code_Segment_Address                   -- keep the code segment address

procedure setup_DFS_critical_error_handler()
    atom save_segment, segment_ad           -- (get code & data segment)
    sequence save_segment_code
    sequence int24_dfs_code                 -- DFS-handler code

    -- allocate space for storing segment address:
    segment_ad = allocate(4)
    if not segment_ad then return end if    -- cannot setup handler
    lock_memory(segment_ad, 4)

    -- 1. save the value of the code segment in Code_Segment_Address:
    save_segment_code = {
	#53,    -- push ebx
	#0E,    -- push cs   or #1E push ds -- only a 16-bit value
	#5B,    -- pop ebx
	#89, #1D} & int_to_bytes(segment_ad) & -- mov segment_ad, ebx
	{#5B,   -- pop ebx
	#C3     -- ret
    }           -- note: there's no "pop cs" instruction

    save_segment = allocate(length(save_segment_code))
    if not save_segment then                -- can't allocate memory?
	free(segment_ad)                    -- free segment_ad buffer
	return                              -- cannot setup handler
    end if
    poke(save_segment, save_segment_code)   -- save code segment
    call(save_segment)                      -- (get low order 2-bytes):
    Code_Segment_Address = peek(segment_ad) + 256 * peek(segment_ad + 1)

    -- 2. load the value of the data segment to segment_ad buffer:
    poke(save_segment + 1, #1E)
    call(save_segment)                      -- save data segment
    free(save_segment)                      -- free save_segment buffer

    -- 3. load DFS-handler code to memory:
    int24_dfs_code = {
	-- save registers:
	#1E,    -- push ds
	#53,    -- push ebx

	-- restore the data segment value (low-order 2 bytes of segment_ad):
	#BB} &
	peek({segment_ad, 2}) & {0, 0} &    -- mov ebx, data segment value
	{#53,   -- push ebx
	#1F} &  -- pop ds

	-- set flag when error occurs (I don't know shorter instruction...):
	#BB & repeat(SET, 4) &              -- mov ebx, non-zero-value
	{#89, #1D} & int_to_bytes(INT24_FLAG_ADDRESS) & -- mov flag_ad, ebx

	-- acknowledge the interrupt in al,
	-- #03=Fail the function call in progress:
	-- * notice: value of #01 (ignore the error) might keep the floppy
	--           motor running forever, etc - so it must *not* be used.
	{#B0, #03} &    -- mov al, #03

	-- restore registers:
	{#5B,   -- pop ebx
	#1F,    -- pop ds

	-- return from interrupt back to DOS (DOS is still allowed to abort
	-- the program in some very critical errors, as memory corruption):
	#CF     -- iretd
    }
    free(segment_ad)                        -- free segment_ad buffer

    INT24_dfs_address = allocate(length(int24_dfs_code))
    if not INT24_dfs_address then           -- can't allocate memory?
	return                              -- cannot setup handler
    end if

    -- load DFS-handler code and lock it forever:
    poke(INT24_dfs_address, int24_dfs_code)
    lock_memory(INT24_dfs_address, length(int24_dfs_code))
end procedure



-- PURPOSE: disable/enable DOS (usual) Critical-Error handler in either
--          real or protected mode (the "Abort, Retry, Ignore, Fail?" message).
-- NOTE:    use check_error() to check if critical error occurred.
-- DETAILS: allow is FALSE/TRUE (disable/enable DOS-critical-error-handler).
-- INPUT:   INT24_usual_address, INT24_dfs_address, Code_Segment_Address
-- OUTPUT   INT24_usual_address
-- RETURN:  boolean; FALSE/TRUE - the previous state of allow_error().
--           * the previous state may be saved and restored after use.
-- EXAMPLE: i = allow_error(FALSE)  --> disable DOS Critical-Error handler
-- SEEALSO: check_error, drive_ready

sequence INT24_usual_address    -- Int #24=Critical-Error handler address
    INT24_usual_address = EMPTY -- EMPTY=DOS-handler is enabled by default

global function allow_error(boolean allow)
    boolean is_enabled

    -- clear error flag since few routines might call allow_error() at
    -- the same time:
    if check_error() then end if

    -- keep previous allow_error() state (disabled/enabled):
    is_enabled = equal(INT24_usual_address, EMPTY) -- EMPTY=DOS handler enabled


    -- enable the DOS critical error handler (only if disabled):
    if allow and not is_enabled then
	set_vector(#24, INT24_usual_address)    -- restore DOS handler
	INT24_usual_address = EMPTY             -- EMPTY=DOS handler enabled

    -- disable the DOS critical error handler (only if enabled):
    elsif not allow and is_enabled then

	-- needs to setup the DFS-handler only once, at first call:
	if not INT24_dfs_address then
	    setup_DFS_critical_error_handler()
	end if

	-- save DOS-handler address (could be also another user-handler),
	-- then set DFS-handler address:
	if INT24_dfs_address then
	    -- (Int #24=Critical-Error handler address).
	    INT24_usual_address = get_vector(#24)
	    set_vector(#24, {Code_Segment_Address, INT24_dfs_address})
	end if
    end if

    -- return the previous state of allow_error():
    return is_enabled
end function



--bookmark: PATH EXISTS ROUTINES; ===========================================

-- PURPOSE: check if drive of pn exists (return TRUE for all mapped-drives).
-- NOTE:    this function cannot cause a critical error.
-- CREDIT:  David Cuny, EE:CS Editor - hard_drives(), for using AX=#4409.
-- RETURN:  boolean; FALSE/TRUE (error/drive exists).
-- EXAMPLE: i = drive_exists("..\\Path Name")   --> check using path
--          i = drive_exists('C')               --> check only drive letter
-- SEEALSO: drives_list, drive_ready, real_drive
global function drive_exists(object pn)
    sequence r                              -- list of register values
    integer drive                           -- drive letter

    if atom(pn) then pn &= ':' end if       -- e.g. 'C' --> "C:"
    pn = trim_path(pn)                      -- trim invalid characters.
    drive = find(':', pn)                   -- find drive delimiter (must be
					    --  position 2 in correct syntax).
    if drive = 2 then                       -- drive specified ("x:...")?
	drive = and_bits(pn[1], UPPER)      -- upper(drive)

	if drive >= 'A' and drive <= 'Z' then   -- valid drive letter?
	    r = repeat(CLEAR, REG_LIST_SIZE) -- initialize registers list
	    r[REG_AX] = #4409               -- AH=#44; AL=#09 - IOCTL: Check
					    --  if Block Device is Remote function
	    r[REG_BX] = drive - '@'         -- BL=drive number (0=default, 1=A, 2=B, etc.)
	    r = dos_interrupt(#21, r)       -- call DOS interrupt #21
	    if not and_bits(r[REG_FLAGS], 1) then   -- Carry Flag set on error
		return TRUE                 -- drive is exist on system.
	    end if
	end if

    elsif not drive then                    -- drive not specified:
	return TRUE                         --  then it's the current drive.
    end if

    return FALSE                            -- FALSE = invalid pn or drive
end function



-- PURPOSE: check if dir pn exists (not root-dir or file).
-- RETURN:  boolean; FALSE/TRUE (error/dir exists).
-- EXAMPLE: i = dir_exists("C:\\Path Name")
-- SEEALSO: file_exists, drive_exists, wdir, real_path
global function dir_exists(sequence pn)
    if find('*', pn) or find('?', pn) then      -- wildcards is not dir:
	return FALSE
    else                                        -- find dirs only:
	return sequence(wdir({pn, 'd'}))        -- wdir() will trim_path().
    end if
end function



-- PURPOSE: check if file pn exists (not dir).
-- RETURN:  boolean; FALSE/TRUE (error/file exists).
-- EXAMPLE: i = file_exists("C:\\Path Name")
-- SEEALSO: file_size, dir_exists, drive_exists, wdir, real_path
global function file_exists(sequence pn)
    if find('*', pn) or find('?', pn) then      -- wildcards is not a file:
	return FALSE
    else                                        -- find files only:
	return sequence(wdir({pn, 'f'}))        -- wdir() will trim_path().
    end if
end function



--bookmark: CURRENT PATH ROUTINES; ==========================================

-- PURPOSE: return the current working drive letter.
-- RETURN:  integer; current drive letter upper cased ('A'..'Z').
-- EXAMPLE: i = curdrive()     --> i may be 'C'
-- SEEALSO: setdrive, curdir
global function curdrive()
    sequence r                              -- list of register values

    r = repeat(CLEAR, REG_LIST_SIZE)        -- initialize registers list
    r[REG_AX] = #1900                       -- AH=#19 - Current Disk function
    r = dos_interrupt(#21, r)               -- call DOS interrupt #21
    return and_bits(r[REG_AX], #FF) + 'A'   -- AL=drive code (0=A, 1=B, etc.)
end function



-- PURPOSE: return the current dir of the current drive.
-- INPUT:   DFS_lfn, CDDrive (output by curdirs()).
-- RETURN:  string; "x:."/current-dir (error/successful).
-- EXAMPLE: s = curdir()   --> current dir on the current drive
-- SEEALSO: curdirs, setdir, curdrive, real_dir
xDrive_letter
    CDDrive                                 -- drive letter to check
    CDDrive = '@'                           -- '@'=check current-drive
global function curdir()
    sequence r                              -- list of register values
    integer ad                              -- low memory address
    sequence pn                             -- current dir
    integer len_pn                          -- buffer length to hold pn
    xDrive_letter drive                     -- drive to check

    drive = CDDrive                         -- don't output to CDDrive; use
					    --  private variable instead.

    if drive = '@' then                     -- replace '@' with the
	drive = curdrive()                  --  current drive letter (upper)
    else
	drive = and_bits(drive, UPPER)      -- upper(drive)
    end if

    r = repeat(CLEAR, REG_LIST_SIZE)        -- initialize registers list
    if DFS_lfn then                         -- Get Current Directory function:
	r[REG_AX] = #7147                   --  AX=#7147 (LFN)
	len_pn = 260 + 1                    --  max pn length (260 usually...)
    else
	r[REG_AX] = #4700                   --  AH=#47   (regular)
	len_pn = 128 + 1                    --  max pn length (128 theoretically)
    end if

    ad = allocate_low(len_pn)               -- low memory to hold ASCIIZ pn
    if not ad then                          -- memory can't be allocated?
	return drive & ":."                 -- ("x:." equals current dir)
    end if

    r[REG_DX] = drive - '@'                 -- DL=drive code (0=default, 1=A, etc.)
    r[REG_DS] = floor(ad / 16)              -- DS:SI=segment:offset for
    r[REG_SI] = remainder(ad, 16)           --  ASCIIZ pn
    r = dos_interrupt(#21, r)               -- call DOS interrupt #21
    pn = peek({ad, len_pn})                 -- get dir from buffer
    free_low(ad)                            -- free pn buffer

    if and_bits(r[REG_FLAGS], 1) then       -- Carry Flag set on error:
	return drive & ":."                 -- ("x:." equals current dir)
    else                                    -- if successful:
	return drive & ":\\" & trim_asciiz(pn)
    end if
end function



-- PURPOSE: return the current dir of drive 'A'..'Z'.
-- OUTPUT:  CDDrive (input by curdir()).
-- RETURN:  string; "x:."/current-dir-of-drive (error/successful).
-- EXAMPLE: s = curdirs('E')   --> current dir of drive E:
-- SEEALSO: curdir, setdir, curdrive
global function curdirs(drive_letter drive)
    sequence cd                             -- current dir

    -- output the specified drive to curdir():
    CDDrive = drive

    -- get current dir on the specified drive:
    cd = curdir()

    -- reset curdir() to check the current drive ('@'=current-drive):
    CDDrive = '@'

    return cd
end function



--bookmark: SET CURRENT PATH ROUTINES; ======================================

-- PURPOSE: get/set logical drive map ('A'..'Z'). similar to DOS's treatment
--          of a single physical floppy drive as both A: and B:.
--          * This routine avoids DOS message to replace a disk in the drive.
-- NOTE:    it use existing mapped-drive-letters (it cannot map new letters).
-- DETAILS: mode is: GET/SET (get/set logical drive map).
-- RETURN:  integer; drive 'A'..'Z' (the last letter used to reference drive).
-- EXAMPLE: i = drive_map('A', GET) --> get the letter used to reference A:
--          i = drive_map('B', SET) --> set the letter to reference B:
-- SEEALSO: setdrive, drive_ready, drives_list
function drive_map(drive_letter drive, boolean mode)
    sequence r                              -- list of register values

    drive = and_bits(drive, UPPER)          -- upper(drive)
    r = repeat(CLEAR, REG_LIST_SIZE)        -- initialize registers list
    r[REG_AX] = #440E + (mode and SET)      -- AH=#44; AL=#0E/#0F - IOCTL:
					    --  Get/Set Logical Drive Map function.
    r[REG_BX] = drive - '@'                 -- BL=drive number (0=default, 1=A, 2=B, etc.)
    r = dos_interrupt(#21, r)               -- call DOS interrupt #21
					    -- AL=mapping code (1=A, 2=B, etc).
    r[REG_AX] = and_bits(r[REG_AX], #FF) + '@'

    if and_bits(r[REG_FLAGS], 1) then       -- Carry Flag set on error
	return drive                        -- error/function not supported.
    elsif r[REG_AX] < 'A' or r[REG_AX] > 'Z' then
	return drive                        -- AL=#00 if drive not mapped.
    else
	return r[REG_AX]                    -- the last letter used for drive
    end if
end function



-- PURPOSE: change the current working drive to 'A'..'Z'.
-- NOTICE:  * it will change even if drive is not ready! but drive must exist.
--          * it automatically set the logical drive map (see drive_map()).
-- EXAMPLE: setdrive('C')   --> C: is now the current/default drive.
-- SEEALSO: curdrive, drives_list, drive_ready, setdir, drive_map
global procedure setdrive(drive_letter drive)
    sequence r                              -- list of register values

    drive = and_bits(drive, UPPER)          -- upper(drive)
    r = repeat(CLEAR, REG_LIST_SIZE)        -- initialize registers list
    r[REG_AX] = #0E00                       -- AH=#0E - Select Drive function
    r[REG_DX] = drive - 'A'                 -- DL=drive code (0=A, 1=B, etc.)
    r = dos_interrupt(#21, r)               -- call DOS interrupt #21
    drive = drive_map(drive, SET)           -- auto-set logical drive map.
end procedure



-- PURPOSE: change the current dir to pn.
-- INPUT:   DFS_lfn
-- RETURN:  boolean; FALSE/TRUE (error/successful).
-- EXAMPLE: i = setdir("C:\\Temp Directory")
-- SEEALSO: curdir, make_dir, setdrive
global function setdir(sequence pn)
    sequence r                              -- list of register values
    integer ad                              -- low memory address

    pn = trim_path(pn) & NULL               -- trim invalid characters.
					    --  NULL = ASCII --> ASCIIZ
    ad = allocate_low(length(pn))           -- low memory buffer to hold pn
    if not ad then return FALSE end if      -- memory can't be allocated?

    poke(ad, pn)                            -- poke ASCIIZ pn to buffer
    r = repeat(CLEAR, REG_LIST_SIZE)        -- initialize registers list
    if DFS_lfn then                         -- Change Directory function:
	r[REG_AX] = #713B                   --  AX=#713B (LFN)
    else
	r[REG_AX] = #3B00                   --  AH=#3B   (regular)
    end if
    r[REG_DS] = floor(ad / 16)              -- DS:DX=segment:offset of
    r[REG_DX] = remainder(ad, 16)           --  ASCIIZ pn
    r = dos_interrupt(#21, r)               -- call DOS interrupt #21
    free_low(ad)                            -- free pn buffer
    return not and_bits(r[REG_FLAGS], 1)    -- Carry Flag set on error
end function



--bookmark: REAL PATH ROUTINES; =============================================

-- PURPOSE: return drive letter of pn. pn may exist or not. drive must exist.
-- NOTE:    this function cannot cause a critical error.
-- RETURN:  integer; FAIL/drive 'A'..'Z' (error/upper-drive-letter).
-- EXAMPLE: i = real_drive("..\\Directory Name") --> i is curdrive (e.g. 'C').
-- SEEALSO: real_dir, real_path, drive_ready, move_file, drive_name
global function real_drive(sequence pn)
    pn = trim_path(pn)                      -- trim invalid characters.

    if drive_exists(pn) then                -- drive exists on system?
	if find(':', pn) = 2 then           -- valid drive specified?
	    return and_bits(pn[1], UPPER)   --  return upper(drive)
	else                                -- drive not specified:
	    return curdrive()               --  then it's the current drive.
	end if
    else
	return FAIL                         -- FAIL = invalid pn or drive
    end if
end function



-- PURPOSE: return the full path of pn. pn is an existing root/sub-dir.
-- NOTE:    use real_path() for any kind of pathname (not only dir).
-- RETURN:  object; integer FAIL/string full-dirname (error/successful).
--           * if pn is empty then returns current-dir string.
-- EXAMPLE: x = real_dir("..\\Directory Name")
-- SEEALSO: real_path, dir_exists, real_short_path, real_drive, dir_name
global function real_dir(sequence pn)
    integer drive                           -- drive of pn
    sequence keepDir                        -- keep current dir of drive
    integer nul                             -- accept and forget

    pn = trim_path(pn)                      -- trim invalid characters.

    if length(pn) then                      -- pn specified?
	drive = real_drive(pn)              --  get real drive of pn
	if drive = FAIL then                --  pn using invalid drive?
	    return FAIL                     --  drive/path is invalid
	else
	    keepDir = curdirs(drive)        --  keep current dir of drive
	end if
    else                                    -- pn not specified:
	return curdir()                     --  the fullname is current dir.
    end if

    -- let DOS resolve the path. simple. accurate:
    if setdir(pn) then                      -- try to change dir to pn:
	pn = curdirs(drive)                 --  get current dir (fullname)
	nul = setdir(keepDir)               --  restore old current dir
	return pn                           --  return fullname
    else                                    -- on error:
	return FAIL                         --  dir invalid/not exist
    end if
end function



-- PURPOSE: return the fullname of pn. the dir portion of pn must exist but
--          the file portion may not exist. wildcards allowed in file portion.
-- DETAILS: mode: FALSE = return string fullname.
--                 TRUE = return sequence {string fullname, integer path-type}.
--                     * path-type is:
--                      'r' = fullname is an existing root-dir.
--                      'd' = fullname is an existing sub-dir.
--                      'f' = fullname is an existing file.
--                      'w' = the file portion has wildcards (* or ?).
--                      'n' = the file portion not exist on system.
-- RETURN:  object; integer FAIL/sequence, see mode (error/successful).
-- EXAMPLE: x = real_path("..\\Pathname\\..\\", FALSE)
-- SEEALSO: real_dir, real_short_path, real_drive, dir_name
global function real_path(sequence pn, boolean mode)
    object info
    sequence file                           -- filename portion
    integer t                               -- path type (r/d/f/w/n)

    pn = trim_path(pn)                      -- trim invalid characters once.

    info = real_dir(dir_name(pn))           -- get full dir portion
    if atom(info) then                      -- dir not exist?
	return FAIL                         --  error
    else                                    -- dir portion exist:
	file = file_name(pn)                --  get filename portion
	if length(file) then                --  is there filename portion?
	    pn = slash_path(info) & file    --  pn is now full path
	else                                --  (pn could be just ".", etc):
	    pn = info                       --  pn is now full dir (no filename)
	end if
	if not mode then                    --  don't get also path-type?
	    return pn                       --  return only fullname
	end if
    end if

    -- get path-type, r/d/f/w/n (root/dir/file/wildcards/not-exist):
    if pn[$] = '\\' then
	t = 'r'                             -- only root-dir specified.
    elsif not length(file) then
	t = 'd'                             -- only sub-dir specified.
    elsif find('*', file) or find('?', file) then
	t = 'w'                             -- file portion has wildcards
    else
	info = wdir(pn)                     -- file/sub-dir exist?

	if atom(info) then
	    t = 'n'                         -- file portion not exist
	elsif find('d', info[1][2]) then    -- ([2]=D_ATTRIBUTES)
	    t = 'd'                         -- file portion is sub-dir
	else
	    t = 'f'                         -- file portion is file
	end if
    end if

    return {pn, t}                          -- return {"pn", 'type'}
end function



-- PURPOSE: return unique name for creating a temporary file or dir in
--          existing directory pn. The name is made of 8 capital letters A..Z.
-- NOTE:    * this name must be used immediatly since file-system may change.
--          * any filename-extension can be safely appended to this name.
-- RETURN:  object; integer FAIL/string fullname (error/successful).
-- EXAMPLE: x = real_temp(".")         --> return name in current-dir
--          x = real_temp("C:\\TEMP")  --> return name in C:\TEMP
-- SEEALSO: open_file, make_dir
global function real_temp(sequence pn)
    sequence randName                       -- filename-string to randomize.
    sequence filename                       -- the temporary filename.
    object info                             -- resolve dir-name

    info = real_dir(pn)                     -- get full directory name.
					    --  real_dir() will trim_path().
    if sequence(info) then
	pn = slash_path(info)               -- add backslash to full dir-name
	randName = repeat('Z' - 'A' + 1, 8) -- filename of 8 A..Z letters.
    else
	return FAIL                         -- directory not exist.
    end if

    -- find a random-unique name:
    for i = 1 to 5000 do                    -- try 5000 names (enough...?)
	filename = pn & (rand(randName) + '@') -- make random name

	-- if base-filename/base-dirname not exist then return this name:
	if atom(wdir(filename & ".*")) then -- use ".*" to find any extension
	    return filename                 -- return full pathname
	end if
    end for
					    -- unique-name not found after
    return FAIL                             --  5000 tries!
end function



--bookmark: DRIVE ROUTINES; =================================================

-- PURPOSE: check if drive of pn is ready for use (avoid a critical error).
-- NOTE:    it automatically set the logical drive map (see drive_map()).
-- RETURN:  boolean; FALSE/TRUE (error/drive is ready).
-- EXAMPLE: i = drive_ready("..\\EUPHORIA\README.DOC") --> check using path
--          i = drive_ready('C')                --> check only drive letter
-- SEEALSO: drives_list, setdrive, real_drive, drive_map, drive_exists
global function drive_ready(object pn)
    integer drive                           -- drive letter to check
    integer status                          -- critical-error current status
    integer error                           -- critical-error occurred?
    integer nul                             -- accept and forget

    if atom(pn) then pn &= ':' end if       -- e.g. 'C' --> "C:"
    drive = real_drive(pn)                  -- get real drive letter

    if drive = FAIL then                    -- drive exists?
	return FALSE                        --  drive not exists (not ready).
    else                                    -- drive exists:
	nul = drive_map(drive, SET)         --  auto-set logical drive map
	status = allow_error(FALSE)         --  disable critical-error handler
	nul = open(drive & ":\\awrgd", "r") --  access the drive - open file.
	error = check_error()               --  critical error occurred?
	if nul != -1 then close(nul) end if --  close file, if actually opened
	nul = allow_error(status)           --  restore previous handler
	return not error                    --  no error? then drive is ready.
    end if
end function



-- PURPOSE: return a list of currently available logical drives in the system.
-- NOTE:    this function cannot cause a critical error.
-- DETAILS: mode: FALSE = add only last letter used by mapped-drives (default).
--                TRUE  = add all letters used by mapped-drives.
-- RETURN:  string; "drives"
-- EXAMPLE: s = drives_list()    --> s may be "ACDG"; "BC"; etc.
-- SEEALSO: setdrive, volume_label, drive_ready, drive_map
global function drives_list(boolean mode)
    sequence list                           -- lists of drives

    list = EMPTY                            -- init drives list

    -- add all logical drives to the list (upper-cased):
    for drive = 'A' to 'Z' do
	if drive_exists(drive) then         -- drive exists on system?
	    -- mode=TRUE, add all letters used by mapped-drives:
	    if mode then
		list &= drive               -- add drive to list

	    -- mode=FALSE, add only last letter used by mapped-drives (default):
	    elsif drive = drive_map(drive, GET) then  -- get logical drive map.
		list &= drive               -- add drive to list
	    end if
	end if
    end for

    return list
end function



-- PURPOSE: return the volume-label of drive 'A'..'Z' (up to 11 characters).
-- RETURN:  string; EMPTY/volume-label (error/successful).
-- EXAMPLE: s = volume_label('C')   --> s may be "FREEDOS2011", etc.
-- SEEALSO: label_drive, wdir, curdrive, drives_list
global function volume_label(drive_letter drive)
    object s                                -- volume label
    integer dot                             -- dot in 8.3 format filename

    -- volume label may exist only in the root dir of drive - find it:
    s = wdir({drive & ":\\*.*", 'v'})

    if sequence(s) then                     -- volume label found:
	s = s[1][1]                         -- ([1]=D_NAME)
	dot = find('.', s)                  -- find dot in label (8.3 format)
	if dot then                         -- remove the dot from label
	    s = s[1..dot - 1] & s[dot + 1..$]
	end if
	return s                            -- return label (11 chars maximum)
    else                                    -- volume label not found:
	return EMPTY                        -- return empty string
    end if
end function



-- PURPOSE: create/change/delete the volume label of drive.
-- DETAILS: label is a 0 to 11 bytes string. if trimmed label is empty then
--          current label is deleted.
-- RETURN:  boolean; FALSE/TRUE (error/successful).
-- EXAMPLE: i = label_drive('A', "BACKUP1")
-- SEEALSO: volume_label, wdir
global function label_drive(drive_letter drive, sequence label)
    sequence r                              -- list of register values
    integer ad_xfcb, ad_fcb                 -- low memory XFCB, FCB addresses
    integer segment_xfcb, offset_xfcb       -- ad_xfcb segment, offset
    object cur_label                        -- get current-label (if exists)
    sequence keepDir                        -- FCB can use only current dir
    integer nul                             -- accept and forget

    drive = and_bits(drive, UPPER)          -- upper(drive)
    label = trim_path(label)                -- trim invalid characters.

    -- verify that drive and label are valid:
    if not drive_exists(drive) or           -- invalid drive letter
	length(label) > 11 or               -- 11 chars is maximum (8 + 3)
	find('.', label) or                 -- don't allow '.'/extension.
	find('*', label) or find('?', label) -- wildcards are not filename.
    then
	return FALSE                        -- invalid drive or label
    end if

    -- common for the FCB functions:
    ad_xfcb = allocate_low(7 + 37)          -- low memory buffer to hold XFCB
    if not ad_xfcb then                     -- memory can't be allocated?
	return FALSE
    else                                    -- according to Ray Duncan:
	mem_set(ad_xfcb, 0, 7 + 37)         -- "Zero out the prospective FCB."
    end if
    ad_fcb = ad_xfcb + #07                  -- FCB address shifted by 7-bytes
    segment_xfcb = floor(ad_xfcb / 16)      -- segment:offset of extended FCB
    offset_xfcb = remainder(ad_xfcb, 16)
    poke(ad_fcb, drive - '@')               -- drive number (0=default, 1=A, etc)
    poke(ad_xfcb, #FF)                      -- #FF signature for extended FCB
    poke(ad_xfcb + #06, #08)                -- file attribute if extended FCB
					    --  (0b_0000_1000 = #08 = volume)
    keepDir = curdirs(drive)                -- FCB can use only current-dir of
    nul = setdir(drive & ":\\")             --  drive (label is in root-dir).
    r = repeat(CLEAR, REG_LIST_SIZE)        -- initialize registers list

    -- find current label ('.'/extension position is unknown):
    cur_label = wdir({drive & ":\\*.*", 'v'})

    -- delete current label:
    if sequence(cur_label) then             -- current label exists?
	cur_label = cur_label[1][1]         -- ([1]=D_NAME)
	mem_set(ad_fcb + #01, 32, 8 + 3)    -- pad filename with spaces
	poke(ad_fcb + #01, filebase_name(cur_label))
	poke(ad_fcb + #09, fileext_name(cur_label))
	r[REG_AX] = #1300                   -- AH=#13 - Delete File using FCB function
	r[REG_DS] = segment_xfcb            -- DS:DX=segment:offset of XFCB
	r[REG_DX] = offset_xfcb
	r = dos_interrupt(#21, r)           -- call DOS interrupt #21 (AL=#00 if successful).
    elsif not length(label) then            -- delete not-existing-label?
	r[REG_AX] = #FF                     -- AL=#FF if failed to delete.
    end if

    -- create new label and close the file (if not error and label supplied):
    if not and_bits(r[REG_AX], #FF) and length(label) then
	mem_set(ad_fcb + #01, 32, 8 + 3)    -- pad filename with spaces
	poke(ad_fcb + #01, label)           -- poke new label (without '.')
	r = r and CLEAR                     -- clear registers list
	r[REG_AX] = #1600                   -- AH=#16 - Create File using FCB function
	r[REG_DS] = segment_xfcb            -- DS:DX=segment:offset of XFCB
	r[REG_DX] = offset_xfcb
	r = dos_interrupt(#21, r)           -- call DOS interrupt #21
	if not and_bits(r[REG_AX], #FF) then -- AL=#00 if successful.
	    r = r and CLEAR                 -- clear registers list
	    r[REG_AX] = #1000               -- AH=#10 - Close File using FCB function
	    r[REG_DS] = segment_xfcb        -- DS:DX=segment:offset of XFCB
	    r[REG_DX] = offset_xfcb
	    r = dos_interrupt(#21, r)       -- call DOS interrupt #21 (AL=#00 if successful).
	end if
    end if

    free_low(ad_xfcb)                       -- free ad_xfcb buffer
    nul = setdir(keepDir)                   -- restore original dir
    nul = open(drive & ":\\", "r")          -- flush buffers to get new label.
    return not and_bits(r[REG_AX], #FF)     -- AL=#00 if successful (all functions).
end function



--bookmark: DIR/FILE ENTRY ROUTINES; ========================================

-- PURPOSE: rename and/or move file on the same drive; or rename dir.
-- DETAILS: pn is file/dir to rename. pn2 is the new file/dir name.
-- INPUT:   DFS_lfn
-- RETURN:  boolean; FALSE/TRUE (error/successful).
-- EXAMPLE: i = rename("C:\\Temp Directory", "C:\\Temporary Dir")
-- SEEALSO: move_file, copy_file, open_file, make_dir, real_temp
global function rename(sequence pn, sequence pn2)
    sequence r                              -- list of register values
    integer ad_pn, ad_pn2                   -- low memory addresses

    pn = trim_path(pn) & NULL               -- trim invalid characters.
    pn2 = trim_path(pn2) & NULL             --  NULL = ASCII --> ASCIIZ

    ad_pn = allocate_low(length(pn))        -- low memory to hold ASCIIZ pn
    if not ad_pn then                       -- memory can't be allocated?
	return FALSE
    else
	ad_pn2 = allocate_low(length(pn2))  -- low memory to hold ASCIIZ pn2
	if not ad_pn2 then                  -- memory can't be allocated?
	    free_low(ad_pn)                 -- free the pn buffer
	    return FALSE
	end if
    end if

    poke(ad_pn, pn)                         -- poke ASCIIZ pn to buffer
    poke(ad_pn2, pn2)                       -- poke ASCIIZ pn2 to buffer
    r = repeat(CLEAR, REG_LIST_SIZE)        -- initialize registers list
    if DFS_lfn then                         -- Rename File or Directory function:
	r[REG_AX] = #7156                   --  AX=#7156 (LFN)
    else
	r[REG_AX] = #5600                   --  AH=#56   (regular)
    end if
    r[REG_DS] = floor(ad_pn / 16)           -- DS:DX=segment:offset for
    r[REG_DX] = remainder(ad_pn, 16)        --  ASCIIZ pn
    r[REG_ES] = floor(ad_pn2 / 16)          -- ES:DI=segment:offset for
    r[REG_DI] = remainder(ad_pn2, 16)       --  ASCIIZ pn2
    r = dos_interrupt(#21, r)               -- call DOS interrupt #21
    free_low(ad_pn)                         -- free pn buffer
    free_low(ad_pn2)                        -- free pn2 buffer
    return not and_bits(r[REG_FLAGS], 1)    -- Carry Flag set on error
end function



-- PURPOSE: close a file (in regular and LFN modes).
-- DETAILS: handle is returned by open_dos_handle() (*not* Euphoria open()).
-- EXAMPLE: close_dos_handle(2)
-- SEEALSO: open_dos_handle, open_file
procedure close_dos_handle(integer handle)
    sequence r                              -- list of register values

    if handle = 0 then return end if        -- handle 0 closes the keyboard!

    r = repeat(CLEAR, REG_LIST_SIZE)        -- initialize registers list
    r[REG_AX] = #3E00                       -- AH=#3E - Close File function
    r[REG_BX] = handle                      -- BX=DOS file handle
    r = dos_interrupt(#21, r)               -- call DOS interrupt #21
					    -- ( Carry Flag set on error )
    -- < the following code is for debugging only >:
    -- printf(1, "\nDEBUG: file handle %d closed? %d\n", {handle, not and_bits(r[REG_FLAGS], 1)})
end procedure



-- PURPOSE: open file pn for read-only to get the DOS file number (handle).
-- NOTE:    close the file with close_dos_handle() (*not* Euphoria close()).
-- RETURN:  integer; FAIL/handle (error/successful).
-- EXAMPLE: i = open_dos_handle("C:\\Temp Directory\\File Name.dat")
-- SEEALSO: date_file, close_dos_handle, open_file
function open_dos_handle(sequence pn)
    sequence r                              -- list of register values
    integer ad                              -- low memory address

    pn &= NULL                              -- ASCII --> ASCIIZ
    ad = allocate_low(length(pn))           -- low memory buffer to hold pn
    if not ad then return FAIL end if       -- memory can't be allocated?

    poke(ad, pn)                            -- poke ASCIIZ pn to buffer
    r = repeat(CLEAR, REG_LIST_SIZE)        -- initialize registers list
 -- if DFS_lfn then --> (note: LFN unused). -- Open File function:
 --     r[REG_AX] = #716C                   --  AX=#716C (LFN)
 --  -- r[REG_BX] = #0000 -- (already 0)    --  BX=#0000 - read-only access mode
 --     r[REG_DX] = #0001                   --  DX=#0001 - Open File sub-function
 --     r[REG_DS] = floor(ad / 16)          --  DS:SI=segment:offset of
 --     r[REG_SI] = remainder(ad, 16)       --   ASCIIZ pn
 -- else
	r[REG_AX] = #3D00                   --  AH=#3D   (regular)
					    --  AL=#00 - read-only access mode
	r[REG_DS] = floor(ad / 16)          --  DS:DX=segment:offset of
	r[REG_DX] = remainder(ad, 16)       --   ASCIIZ pn
 -- end if
    r = dos_interrupt(#21, r)               -- call DOS interrupt #21
    free_low(ad)                            -- free pn buffer

    if and_bits(r[REG_FLAGS], 1) then       -- Carry Flag set on error:
	return FAIL
    else                                    -- successful:
	return r[REG_AX]                    -- AX=DOS file handle
    end if
end function



-- PURPOSE: set the last write date/time of the file pn.
-- DETAILS: dt is the new date-time sequence in dir()'s format:
--           {D_YEAR, D_MONTH, D_DAY, D_HOUR, D_MINUTE, D_SECOND}
-- INPUT:   DFS_lfn
-- RETURN:  boolean; FALSE/TRUE (error/successful).
-- EXAMPLE: i = date_file("C:\\File Name.txt", {2009,6,3, 10,20,55})
-- SEEALSO: attrib, wdir, rename, set_date
global function date_file(sequence pn, sequence dt)
    sequence r                              -- list of register values
    integer ad                              -- low memory address (LFN)
    integer handle                          -- opened file handle (file number)

    pn = trim_path(pn)                      -- trim invalid characters.

    -- convert {D_YEAR, D_MONTH, D_DAY}, {D_HOUR, D_MINUTE, D_SECOND} to
    -- {date, time} words in DOS file-entry format:
    dt = sequence_to_entry_date_time(dt[1..3], dt[4..6])

    r = repeat(CLEAR, REG_LIST_SIZE)        -- initialize registers list
    if DFS_lfn then                         -- Extended Get/Set File Attributes function:
	pn &= NULL                          --  ASCII --> ASCIIZ
	ad = allocate_low(length(pn))       --  low memory buffer to hold pn
	if not ad then return FALSE end if  --  memory can't be allocated?
	poke(ad, pn)                        --  poke ASCIIZ pn to buffer

	r[REG_AX] = #7143                   --  AX=#7143 (LFN)
	r[REG_BX] = #0003                   --  BL=#03 - set last write date/time
	r[REG_CX] = dt[2]                   --  CX=time word
	r[REG_DI] = dt[1]                   --  DI=date word
	r[REG_DS] = floor(ad / 16)          --  DS:DX=segment:offset of
	r[REG_DX] = remainder(ad, 16)       --   ASCIIZ pn

    else                                    -- Set File Date and Time function:
	handle = open_dos_handle(pn)        --  open file to get its handle
	if handle = FAIL then               --  cannot open file?
	    return FALSE
	end if
					    --  AH=#57   (regular)
	r[REG_AX] = #5701                   --  AL=#01 - set date and time
	r[REG_BX] = handle                  --  BX=handle (DOS file number)
	r[REG_CX] = dt[2]                   --  CX=time word
	r[REG_DX] = dt[1]                   --  DX=date word
    end if

    r = dos_interrupt(#21, r)               -- call DOS interrupt #21

    if DFS_lfn then
	free_low(ad)                        -- free pn buffer
    else
	close_dos_handle(handle)            -- close opened file
    end if

    return not and_bits(r[REG_FLAGS], 1)    -- Carry Flag set on error
end function



-- PURPOSE: set file/dir pn attributes (read-only, hidden, system, archive).
-- DETAILS: rhsa (read-only, hidden, system, archive) is the new attributes:
--              "+r" or "-r" = set or clear read-only attribute.
--              "+h" or "-h" = set or clear hidden attribute.
--              "+s" or "-s" = set or clear system attribute.
--              "+a" or "-a" = set or clear archive attribute.
--              (any other strings are invalid and ignored).
--           * volume & directory (v & d) attributes cannot be changed.
-- INPUT:   DFS_lfn
-- RETURN:  boolean; FALSE/TRUE (error/successful).
-- EXAMPLE: i = attrib("C:\\Temp Directory", "+r-h")
--            --> set read-only, clear hidden; system & archive are preserved.
-- SEEALSO: date_file, wdir, entry_attrib_to_string
constant                                    -- masks to set bits 0-5 fast:
     -- ATTR_MASK = {{#01, #02, #04, #08, #10, #20} , "rhsvda"},
    ATTR_SET_MASK = ATTR_MASK[1],           -- set bits mask   (e.g. 0b000001)
  ATTR_CLEAR_MASK = not_bits(ATTR_SET_MASK) -- clear bits mask (e.g. 0b111110)
global function attrib(sequence pn, sequence rhsa)
    sequence r                              -- list of register values
    integer ad                              -- low memory address
    integer attr                            -- new Attributes-Byte

    pn = trim_path(pn) & NULL               -- trim invalid characters.
					    --  NULL = ASCII --> ASCIIZ
    ad = allocate_low(length(pn))           -- low memory buffer to hold pn
    if not ad then return FALSE end if      -- memory can't be allocated?
    poke(ad, pn)                            -- poke ASCIIZ pn to buffer

    -- 1. get current attributes:
    r = repeat(CLEAR, REG_LIST_SIZE)        -- initialize registers list
    if DFS_lfn then                         -- Extended Get/Set File Attributes function:
	r[REG_AX] = #7143                   --  AX=#7143 (LFN)
     -- r[REG_BX] = #0000 -- (already 0)    --  BL=#00 - get attributes
    else                                    -- Get/Set File Attributes function:
	r[REG_AX] = #4300                   --  AH=#43   (regular)
					    --  AL=#00 - get attributes
    end if
    r[REG_DS] = floor(ad / 16)              -- DS:DX=segment:offset of
    r[REG_DX] = remainder(ad, 16)           --  ASCIIZ pn
    r = dos_interrupt(#21, r)               -- call DOS interrupt #21
    if and_bits(r[REG_FLAGS], 1) then       -- Carry Flag set on error
	free_low(ad)                        -- free pn buffer
	return FALSE
    else
	attr = r[REG_CX]                    -- CX=current attributes
    end if

    -- 2. mask current attributes with new attributes:
    rhsa = "-v-d" & rhsa                    -- volume & dir attributes must be
    for i = 1 to 6 do                       --  cleared first (cannot changed).
	-- first clear attribute-bit (*including volume & dir bits*):
	if match('-' & ATTR_MASK[2][i], rhsa) then      --| Attributes-Byte
	    attr = and_bits(attr, ATTR_CLEAR_MASK[i])   --| bits (DOS format):
	-- set attribute-bit:                           --| ------------------
	elsif match('+' & ATTR_MASK[2][i], rhsa) then   --| 0 = 'r' read-only
	    attr = or_bits(attr, ATTR_SET_MASK[i])      --| 1 = 'h' hidden
	end if                                          --| 2 = 's' system
    end for                                             --| 3 = 'v' volume label
							--| 4 = 'd' directory
							--| 5 = 'a' archive
    -- 3. set new attributes:                           --|-------------------
    r = r and CLEAR                         -- clear registers list
    if DFS_lfn then                         -- Extended Get/Set File Attributes function:
	r[REG_AX] = #7143                   --  AX=#7143 (LFN)
	r[REG_BX] = #0001                   --  BL=#01 - set attributes
    else                                    -- Get/Set File Attributes function:
	r[REG_AX] = #4301                   --  AH=#43   (regular)
					    --  AL=#01 - set attributes
    end if
    r[REG_CX] = attr                        -- CX=new attributes
    r[REG_DS] = floor(ad / 16)              -- DS:DX=segment:offset of
    r[REG_DX] = remainder(ad, 16)           --  ASCIIZ pn
    r = dos_interrupt(#21, r)               -- call DOS interrupt #21
    free_low(ad)                            -- free pn buffer
    return not and_bits(r[REG_FLAGS], 1)    -- Carry Flag set on error
end function



--bookmark: DIR/FILE MAIN ROUTINES; =========================================

-- PURPOSE: create the dir pn.
-- INPUT:   DFS_lfn
-- RETURN:  boolean; FALSE/TRUE (error/successful).
-- EXAMPLE: i = make_dir("C:\\Temp Directory")
-- SEEALSO: remove_dir, rename, real_temp, open_file
global function make_dir(sequence pn)
    sequence r                              -- list of register values
    integer ad                              -- low memory address

    pn = trim_path(pn) & NULL               -- trim invalid characters.
					    --  NULL = ASCII --> ASCIIZ
    ad = allocate_low(length(pn))           -- low memory buffer to hold pn
    if not ad then return FALSE end if      -- memory can't be allocated?

    poke(ad, pn)                            -- poke ASCIIZ pn to buffer
    r = repeat(CLEAR, REG_LIST_SIZE)        -- initialize registers list
    if DFS_lfn then                         -- Make Directory function:
	r[REG_AX] = #7139                   --  AX=#7139 (LFN)
    else
	r[REG_AX] = #3900                   --  AH=#39   (regular)
    end if
    r[REG_DS] = floor(ad / 16)              -- DS:DX=segment:offset of
    r[REG_DX] = remainder(ad, 16)           --  ASCIIZ pn
    r = dos_interrupt(#21, r)               -- call DOS interrupt #21
    free_low(ad)                            -- free pn buffer
    return not and_bits(r[REG_FLAGS], 1)    -- Carry Flag set on error
end function



-- PURPOSE: delete the dir pn.
-- NOTE:    cannot delete current-dir, or read-only dir, or dir with files.
-- INPUT:   DFS_lfn
-- RETURN:  boolean; FALSE/TRUE (error/successful).
-- EXAMPLE: i = remove_dir("C:\\Temp Directory")
-- SEEALSO: make_dir, erase_file, rename, attrib
global function remove_dir(sequence pn)
    sequence r                              -- list of register values
    integer ad                              -- low memory address

    pn = trim_path(pn) & NULL               -- trim invalid characters.
					    --  NULL = ASCII --> ASCIIZ
    ad = allocate_low(length(pn))           -- low memory buffer to hold pn
    if not ad then return FALSE end if      -- memory can't be allocated?

    poke(ad, pn)                            -- poke ASCIIZ pn to buffer
    r = repeat(CLEAR, REG_LIST_SIZE)        -- initialize registers list
    if DFS_lfn then                         -- Remove Directory function:
	r[REG_AX] = #713A                   --  AX=#713A (LFN)
    else
	r[REG_AX] = #3A00                   --  AH=#3A   (regular)
    end if
    r[REG_DS] = floor(ad / 16)              -- DS:DX=segment:offset of
    r[REG_DX] = remainder(ad, 16)           --  ASCIIZ pn
    r = dos_interrupt(#21, r)               -- call DOS interrupt #21
    free_low(ad)                            -- free pn buffer
    return not and_bits(r[REG_FLAGS], 1)    -- Carry Flag set on error
end function



-- PURPOSE: delete the file pn (its actual data is not overwritten).
-- NOTE:    cannot delete read-only file.
-- INPUT:   DFS_lfn
-- RETURN:  boolean; FALSE/TRUE (error/successful).
-- EXAMPLE: i = erase_file("C:\\Temp Directory\\File Name.txt")
-- SEEALSO: remove_dir, rename, open_file, attrib
global function erase_file(sequence pn)
    sequence r                              -- list of register values
    integer ad                              -- low memory address

    pn = trim_path(pn) & NULL               -- trim invalid characters.
					    --  NULL = ASCII --> ASCIIZ
    ad = allocate_low(length(pn))           -- low memory buffer to hold pn
    if not ad then return FALSE end if      -- memory can't be allocated?

    poke(ad, pn)                            -- poke ASCIIZ pn to buffer
    r = repeat(CLEAR, REG_LIST_SIZE)        -- initialize registers list
    if DFS_lfn then                         -- Delete File function:
	r[REG_AX] = #7141                   --  AX=#7141 (LFN)
     -- r[REG_SI] = #0000 -- (already 0)    --  SI=#0000 wildcards not allowed
					    --   & search attributes (CX) ignored.
    else
	r[REG_AX] = #4100                   --  AH=#41   (regular)
    end if
    r[REG_DS] = floor(ad / 16)              -- DS:DX=segment:offset of
    r[REG_DX] = remainder(ad, 16)           --  ASCIIZ pn
    r = dos_interrupt(#21, r)               -- call DOS interrupt #21
    free_low(ad)                            -- free pn buffer
    return not and_bits(r[REG_FLAGS], 1)    -- Carry Flag set on error
end function



-- PURPOSE: open file/device to get its Euphoria's file number (see open()).
-- NOTE:    * use close() to close the file (*not* close_dos_handle()).
--          * open()/open_file() supports files up to 2 Gb in size. beyond
--            that some file operations may not work correctly.
-- DETAILS: pn and mode are identical to Euphoria's open() command.
-- INPUT:   DFS_lfn
-- RETURN:  integer; FAIL/Euphoria-file-number (error/successful).
-- EXAMPLE: fn = open_file("C:\\Temp Directory\\File Name.dat", "ab")
-- SEEALSO: erase_file, rename, real_temp, make_dir
global function open_file(sequence pn, sequence mode)
    sequence r                              -- list of register values
    integer ad                              -- low memory address

    pn = trim_path(pn)                      -- trim invalid characters.

    -- LFN files opened for write or append will be created if necessary:
    if DFS_lfn and find(mode, {"w", "wb", "a", "ab"}) then
	ad = allocate_low(length(pn) + 1)   -- low memory buffer to hold pn
	if not ad then return FAIL end if   -- memory can't be allocated?

	poke(ad, pn & NULL)                 -- poke ASCIIZ pn to buffer
	r = repeat(CLEAR, REG_LIST_SIZE)    -- initialize registers list
	r[REG_AX] = #716C                   -- AX=#716C (LFN) - Create/Open File function
     -- r[REG_BX] = #0000 -- (already 0)    -- BX=#0000 - read-only access mode
     -- r[REG_CX] = #0000 -- (already 0)    -- CX=#0000 - create normal attributes
	r[REG_DX] = #0010                   -- DX=#0010 - Create File (fail if exists)
	r[REG_DS] = floor(ad / 16)          -- DS:SI=segment:offset of
	r[REG_SI] = remainder(ad, 16)       --  ASCIIZ pn
	r = dos_interrupt(#21, r)           -- call DOS interrupt #21
	free_low(ad)                        -- free pn buffer
	if not and_bits(r[REG_FLAGS], 1) then -- Carry Flag set on error
	    close_dos_handle(r[REG_AX])     -- AX=DOS file handle (close file)
	end if
    end if

    -- note: there is no point to convert pn to short alias in LFN mode, since
    --       open() opens an *existing* long-file-name in *all* open modes.

    return open(pn, mode)                   -- return regular open() command
end function



--bookmark: DIR/FILE EXTENDED ROUTINES; =====================================

-- PURPOSE: return the size of file pn in bytes.
-- RETURN:  atom; FAIL/size (error/size of file).
-- EXAMPLE: a = file_size("C:\\EUPHORIA\\README.DOC")
-- SEEALSO: file_exists, wdir, real_path
global function file_size(sequence pn)
    object info

    -- get a single file info (don't allow wildcards):
    if not (find('*', pn) or find('?', pn)) then
	info = wdir({pn, 'f'})              -- wdir() will trim_path().
	if sequence(info) then
	    return info[1][3]               -- return [3]=D_SIZE
	end if
    end if

    return FAIL                             -- error/file not exist
end function



-- PURPOSE: compare pn and pn2 files content to check if it's identical.
-- RETURN:  boolean; FALSE/TRUE (error-or-different/successful-no difference).
-- EXAMPLE: i = comp_file("C:\\Work Dir\\My File.TXT", "D:\\My File.BAK")
-- SEEALSO: open_file, copy_file, rename, file_checksum
global function comp_file(sequence pn, sequence pn2)
    integer fn_pn, fn_pn2                   -- file numbers
    atom size_pn, size_pn2                  -- file size
    boolean no_difference                   -- flag (TRUE=no difference)

    pn = trim_path(pn)                      -- trim invalid characters once.
    pn2 = trim_path(pn2)

    -- quickly compare the size of the two files:
    size_pn = file_size(pn)                 -- get size of pn
    if size_pn = FAIL then                  -- file not exists?
	return FALSE                        -- error
    else
	size_pn2 = file_size(pn2)           -- get size of pn2
	if size_pn2 != size_pn then         -- file not exists (FAIL)? or
	    return FALSE                    --  file size is different?
	end if                              -- error or files are different
    end if

    -- open pn and pn2 files in binary mode:
    fn_pn = open_file(pn, "rb")             -- open binary file for reading
    if fn_pn = FAIL then                    -- cannot open file?
	return FALSE                        -- error
    else
	fn_pn2 = open_file(pn2, "rb")       -- open binary file for reading
	if fn_pn2 = FAIL then               -- cannot open file?
	    close(fn_pn)                    -- close pn file
	    return FALSE                    -- error
	end if
    end if

    -- compare files byte by byte and close the files:
    no_difference = TRUE                    -- TRUE=no difference yet
    for i = 1 to size_pn do
	if getc(fn_pn) != getc(fn_pn2) then
	    no_difference = FALSE           -- FALSE=files are different
	    exit
	end if
    end for
    close(fn_pn)                            -- close pn file
    close(fn_pn2)                           -- close pn2 file

    return no_difference                    -- return FALSE or TRUE
end function



-- PURPOSE: copy the file pn to a *new* file pn2 (make identical copy).
-- DETAILS: verify is FALSE/TRUE (don't compare/compare after copy).
-- RETURN:  boolean; FALSE/TRUE (error/successful).
-- EXAMPLE: i = copy_file("C:\\Work Dir\\My File.TXT", "D:\\My File.BAK")
-- SEEALSO: open_file, rename, move_file, comp_file, make_dir, file_checksum
global function copy_file(sequence pn, sequence pn2, boolean verify)
    integer fn_pn, fn_pn2                   -- file numbers
    integer byte                            -- copy one byte at a time.

    pn = trim_path(pn)                      -- trim invalid characters once.
    pn2 = trim_path(pn2)

    if file_exists(pn2) then                -- pn2 file already exists?
	return FALSE                        -- don't overwrite pn2 file (or
    end if                                  --  copy file to itself...).

    -- open pn and pn2 files in binary mode:
    fn_pn = open_file(pn, "rb")             -- open binary file for reading
    if fn_pn = FAIL then                    -- cannot open file?
	return FALSE                        -- error
    else
	fn_pn2 = open_file(pn2, "wb")       -- create binary file for writing
	if fn_pn2 = FAIL then               -- cannot create file?
	    close(fn_pn)                    -- close pn file
	    return FALSE                    -- error
	end if
    end if

    -- copy file then close & flush the files:
    byte = getc(fn_pn)                      -- get first byte from pn file
    while byte != -1 do                     -- while not end of pn file
	puts(fn_pn2, byte)                  -- copy this byte to pn2 file
	byte = getc(fn_pn)                  -- get next byte from pn file
    end while
    close(fn_pn)                            -- close pn file
    close(fn_pn2)                           -- close pn2 file

    -- verify the copy or return successful:
    if verify then
	return comp_file(pn, pn2)           -- copied correctly? (FALSE/TRUE)
    else
	return TRUE                         -- copy successful
    end if
end function



-- PURPOSE: move the file pn to a *new* file-location pn2 (or rename dir).
-- DETAILS: verify is FALSE/TRUE (don't compare/compare after copy).
-- RETURN:  boolean; FALSE/TRUE (error/successful).
-- EXAMPLE: i = move_file("My File.TXT", "D:\\My File.BAK")
-- SEEALSO: rename, copy_file, comp_file, erase_file, file_checksum
global function move_file(sequence pn, sequence pn2, boolean verify)

    pn = trim_path(pn)                      -- trim invalid characters once.
    pn2 = trim_path(pn2)

    -- note: rename() and copy_file() will *not* overwrite pn2 if exists.

    -- move file fast on the same logical drive (no need to verify
    -- since it only updates the dir-entry of the file):
    if real_drive(pn) = real_drive(pn2) then
	return rename(pn, pn2)              -- rename successful (FALSE/TRUE)

    -- copy file, verify if needed, then delete the original file (slower):
    else
	if copy_file(pn, pn2, verify) then
	    return erase_file(pn)           -- delete successful (FALSE/TRUE)
	else
	    return FALSE                    -- error in copy or verify
	end if
    end if
end function



--bookmark: FILE CHECKSUM ROUTINES; =========================================

-- --------------------------- CRC-32 CREDIT: -------------------------------
-- "A PAINLESS GUIDE TO CRC ERROR DETECTION ALGORITHMS" by Ross N. Williams,
-- (ftp.adelaide.edu.au/pub/rocksoft/crc_v3.txt), should explain the CRC-32
-- algorithm which is used here.
-- But instead of fully studying the algorithm, I took the code from "CRC.E",
-- by G.Burke, and rearranged it to use only the default CRC-32 parameters.
-- See CRC.E, by G.Burke, Euphoria archive, for more options.
-- --------------------------------------------------------------------------

-- PURPOSE: generate a constant lookup table for the CRC-32 poly #04C11DB7.
-- NOTE:    function is using CRC-32 Specifications (used also by pkzip).
-- RETURN:  sequence; lookup table of 256 4-bytes-entries (1024-bytes).
-- EXAMPLE: s = CRC32_lookup_table()
-- SEEALSO: CRC32_reflect_table, file_checksum
function CRC32_lookup_table()
    sequence lookup_table                   -- (256 * 4) bytes lookup table
    sequence buffer                         -- temporary buffer of 5-bytes
    sequence poly                           -- #04C11DB7 = CRC-32 default poly
    sequence shifted_poly                   -- mini table of left-shifted poly
    integer shift, top_bit

    lookup_table = repeat({}, 256)          -- container for table

    -- #04C11DB7 = CRC-32 default poly (32-bits width).
    -- prepend implicit '1' bit (have to add another byte -> 00000001):
    poly = 1 & {#04,#C1,#1D,#B7}

    -- make mini table of left-shifted-poly values,
    -- i.e. string of bytes equal to left-shifted-poly from 0 to 7 times:
    shifted_poly = repeat(poly, 8)
    for i = 2 to 8 do
	-- shift a sequence of bytes left by one bit:
	shifted_poly[i] = and_bits(shifted_poly[i - 1] * 2, #FF) +
			((and_bits(shifted_poly[i - 1][2..5], #80) != 0) & 0)
    end for

    -- create lookup table for each possible 8-bit value:
    for i = 0 to 255 do
	-- make a buffer 4 + 1 long, put i in top byte:
	buffer = i & {0,0,0,0}

	-- while any '1' bits remain in the top byte of the buffer
	-- XOR the whole buffer with the appropriate shifted poly
	-- from the mini table above:

	-- return index to leftmost '1' bit in i:
	shift = i
	top_bit = 0
	while shift do
	    top_bit += 1
	    shift = floor(shift / 2)
	end while
	shift = top_bit                     -- shift = leftmost '1' bit in i

	while shift do
	    buffer = xor_bits(buffer, shifted_poly[shift])

	    -- return index to leftmost '1' bit in buffer[1]:
	    shift = buffer[1]
	    top_bit = 0
	    while shift do
		top_bit += 1
		shift = floor(shift / 2)
	    end while
	    shift = top_bit         -- shift = leftmost '1' bit in buffer[1]
	end while

	-- table entry value (4-bytes) is in remainder of buffer:
	lookup_table[i + 1] = buffer[2..5]
    end for

    return lookup_table                     -- return table
end function



-- PURPOSE: generate a constant reflect table for reflecting bytes.
-- NOTE:    function is using CRC-32 Specifications (used also by pkzip).
-- RETURN:  string; 256-bytes reflect table.
-- EXAMPLE: s = CRC32_reflect_table()
-- SEEALSO: CRC32_lookup_table, file_checksum
function CRC32_reflect_table()
    sequence reflect_table                  -- 256-bytes reflect table
    sequence ref_in, ref_out                -- helpers for building table
    integer byte

    reflect_table = repeat(0, 255)
    ref_in =  {128,64,32,16,8,4,2,1}
    ref_out = {1,2,4,8,16,32,64,128}

    -- create table:
    for i = 1 to 255 do
	if not reflect_table[i] then
	    byte = 0
	    for bit = 1 to 8 do             -- reflect a byte (8-bits).
		byte += (ref_out[bit] * (and_bits(i, ref_in[bit]) != 0))
	    end for

	    reflect_table[i] = byte
	    reflect_table[reflect_table[i]] = i
	end if
    end for

    return 0 & reflect_table                -- return 256-bytes string.
end function



-- PURPOSE: read all bytes from a file and return the checksum of its data.
-- NOTE:    * function is using CRC-32 Specifications (used also by pkzip).
--          * test with: #CBF43926 is check value for the string "123456789".
-- RETURN:  object; sequence EMPTY if error.
--                  atom 32-bit CRC checksum of pn if successful.
-- EXAMPLE: x = file_checksum("DFS.E")
-- SEEALSO: comp_file, file_size, wdir
sequence
    LookupTable,                            -- 1024-bytes (256*4) lookup table
    ReflectTable                            -- 256-bytes reflect table
    LookupTable = EMPTY                     -- load tables only on first call
					    --  to speedup multiple checksums.
global function file_checksum(sequence pn)
    integer fn, fileSize                    -- file number, file length
    sequence crc, ref_crc                   -- 4-bytes sequences

    pn = trim_path(pn)                      -- trim invalid characters.

    fileSize = file_size(pn)                -- get file size (length)
    if fileSize = FAIL then
	return EMPTY                        -- EMPTY = error
    end if

    fn = open_file(pn, "rb")                -- open binary file for reading
    if fn = FAIL then return EMPTY end if   -- EMPTY = error

    if not length(LookupTable) then         -- load tables only once:
	LookupTable = CRC32_lookup_table()  -- create 1024-bytes lookup table
	ReflectTable = CRC32_reflect_table() -- create 256-bytes reflect table
    end if

    -- CRC-32 Initial register value is #FFFFFFFF:
    crc = {#FF,#FF,#FF,#FF}

    -- get bytes from file. CRC-32 Reflects the Input (4-bytes sequence):
    for bytes = 1 to fileSize do
	crc = xor_bits(crc[2..4] & 0,
	    LookupTable[xor_bits(crc[1], ReflectTable[getc(fn) + 1]) + 1]
	)
    end for
    close(fn)                               -- close file.

    -- CRC-32 Reflects the Output (4-bytes sequence):
    ref_crc = EMPTY
    for i = 1 to 4 do
	ref_crc = ReflectTable[crc[i] + 1] & ref_crc
    end for

    -- CRC-32 XOR the Out Value with #FFFFFFFF:
    crc = xor_bits(ref_crc, {#FF,#FF,#FF,#FF})

    -- return CRC-32 checksum (bytes in reverse order for bytes_to_int()):
    return bytes_to_int({crc[4], crc[3], crc[2], crc[1]})
end function



--bookmark: SYSTEM CLOCK ROUTINES; ==========================================

-- PURPOSE: set the system time.
-- DETAILS: t is {hour, minute, second}
--           (hour 0..23; minute 0..59; second 0..59).
-- RETURN:  boolean; FALSE/TRUE (error/successful).
-- EXAMPLE: i = set_time({10, 45, 3})
-- SEEALSO: set_date, date_file
global function set_time(sequence t)
    sequence r                              -- list of register values

    r = repeat(CLEAR, REG_LIST_SIZE)        -- initialize registers list
    r[REG_AX] = #2D00                       -- AH=#2D - Set System Time function
    r[REG_CX] = (t[1] * #100) + t[2]        -- CH=hour; CL=minute.
    r[REG_DX] = t[3] * #100                 -- DH=second; (DL=1/100 seconds).
    r = dos_interrupt(#21, r)               -- call DOS interrupt #21
    return and_bits(r[REG_AX], #FF) = #00   -- AL=#00 if successful
end function



-- PURPOSE: set the system date.
-- DETAILS: d is {year, month, day} (year 1980..2099; month 1..12; day 1..31).
-- RETURN:  boolean; FALSE/TRUE (error/successful).
-- EXAMPLE: i = set_date({1997, 6, 21})
-- SEEALSO: set_time, date_file
global function set_date(sequence d)
    sequence r                              -- list of register values

    r = repeat(CLEAR, REG_LIST_SIZE)        -- initialize registers list
    r[REG_AX] = #2B00                       -- AH=#2B - Set System Date function
    r[REG_CX] = d[1]                        -- CX=year.
    r[REG_DX] = (d[2] * #100) + d[3]        -- DH=month; DL=day.
    r = dos_interrupt(#21, r)               -- call DOS interrupt #21
    return and_bits(r[REG_AX], #FF) = #00   -- AL=#00 if successful
end function



-- End of file.
